import 'dart:async';
import 'dart:io';
import 'dart:typed_data';
import 'dart:ui' as ui show Codec;
import 'package:flutter/foundation.dart';
import 'package:flutter/painting.dart';
import 'package:path/path.dart';
import '../flutter_mylibrary.dart';
import 'image_api.dart';
import 'package:path_provider/path_provider.dart';



class NetworkImageWithoutAuth extends ImageProvider<NetworkImageWithoutAuth> with ExtendedImageProvider{


  NetworkImageWithoutAuth(this.url, {this.timeLimit, this.retries, this.timeRetry,
    this.cache, this.scale, this.headers, this.cancelToken});

  ///time limit to request image
  final Duration timeLimit;

  ///the time to retry to request
  final int retries;

  ///the time duration to retry to request
  final Duration timeRetry;

  ///whether cache image to local
  final bool cache;

  /// The URL from which the image will be fetched.
  final String url;

  /// The scale to place in the [ImageInfo] object of the image.
  final double scale;

  /// The HTTP headers that will be used with [HttpClient.get] to fetch image from network.
  final Map<String, String> headers;

  ///token to cancel network request
  final CancellationToken cancelToken;


  @override
  Future<NetworkImageWithoutAuth> obtainKey(ImageConfiguration configuration) {
    return SynchronousFuture<NetworkImageWithoutAuth>(this);
  }



  @override
  ImageStreamCompleter load(NetworkImageWithoutAuth key, Future<ui.Codec> Function(Uint8List bytes, {bool allowUpscaling, int cacheHeight, int cacheWidth}) decode) {

    return MultiFrameImageStreamCompleter(
      codec: _loadAsync(
        key,
        decode,
      ),
      scale: key.scale,
      informationCollector: () {
        return <DiagnosticsNode>[
          DiagnosticsProperty<ImageProvider>('Image provider', this),
          DiagnosticsProperty<NetworkImageWithoutAuth>(
              'Image key', key),
        ];
      },
    );
  }


  Future<ui.Codec> _loadAsync(NetworkImageWithoutAuth key, DecoderCallback decode) async {
    assert(key == this);
    final String md5Key = keyToMd5(key.url);
    ui.Codec result;
    if (cache) {
      try {
        final Uint8List data = await _loadCache(
          key,
          md5Key,
        );
        if (data != null) {
          result = await instantiateImageCodec(data, decode);
        }
      } catch (e) {
        print(e);
      }
    }

    if (result == null) {
      try {
        final Uint8List data = await _loadNetwork(key);
        if (data != null) {
          result = await instantiateImageCodec(data, decode);
        }
      } catch (e) {
        print(e);
      }
    }

    //Failed to load
    if (result == null) {
      //result = await ui.instantiateImageCodec(kTransparentImage);
      return Future<ui.Codec>.error(StateError('Failed to load $url.'));
    }
  }



  ///get the image from cache folder.
  Future<Uint8List> _loadCache(
      NetworkImageWithoutAuth key,
      String md5Key,
      ) async {
    final Directory _cacheImagesDirectory = Directory(
        join((await getTemporaryDirectory()).path, cacheImageFolderName));
    //exist, try to find cache image file
    if (_cacheImagesDirectory.existsSync()) {
      final File cacheFlie = File(join(_cacheImagesDirectory.path, md5Key));
      if (cacheFlie.existsSync()) {
        return await cacheFlie.readAsBytes();
      }
    }
    //create folder
    else {
      await _cacheImagesDirectory.create();
    }
    //load from network
    final Uint8List data = await _loadNetwork(
      key,
    );
    if (data != null) {
      //cache image file
      await File(join(_cacheImagesDirectory.path, md5Key)).writeAsBytes(data);
      return data;
    }

    return null;
  }


  /// get the image from network.
  Future<Uint8List> _loadNetwork(NetworkImageWithoutAuth key) async {
    try {
      final Uint8List bytes = await _tryGetResponse(key.url);

      if (bytes.lengthInBytes == 0) {
        return Future<Uint8List>.error(
            StateError('NetworkImage is an empty file: ${Uri.base.resolve(key.url)}'));
      }
      return bytes;
    } on OperationCanceledError catch (_) {
      print('User cancel request $url.');
      return Future<Uint8List>.error(StateError('User cancel request $url.'));
    } catch (e) {
      print(e);
    } finally {

    }
    return null;
  }



  //http get with cancel, delay try again
  Future<Uint8List> _tryGetResponse(String url) async {
    cancelToken?.throwIfCancellationRequested();
    return await RetryHelper.tryRun<Uint8List>(
          () {
        return CancellationTokenSource.register(
          cancelToken,
          ImageApi.image(url, headers, timeLimit),
        );
      },
      cancelToken: cancelToken,
      timeRetry: timeRetry,
      retries: retries,
    );
  }

}
